// Author: Pete Kirkham
//
// quad store.
//
// simple file-reader

#ifndef _FILE_READER_H_
#define _FILE_READER_H_

#include "io/reader_base.h"
//#include "utils.h"

#define _XOPEN_SOURCE 600
#include <fcntl.h>

class file_reader
{
public:
    inline file_reader (int file) : _file(file), _block_start(-BLOCK_SIZE), _fail(false), _index(BLOCK_SIZE)
    {
        _out_buf[0] = 0;
    //    std::cout << "reader (" << to_hex(this) << "," << _file << "," << _index << ")" << std::endl;
    }

    inline size_t  total_read ()
    {
        return _block_start + _index;
    }

    inline bool    failed ()
    {
        return _fail;
    }

    inline bool    ok ()
    {
        return !_fail;
    }

    inline bool    seek (int offset)
    {
        if (lseek(_file, offset, SEEK_SET) != (off_t)-1)
        {
            _block_start = offset;
            _index = BLOCK_SIZE;

            return true;
        }
        else
        {
            return false;
        }
    }

    inline const char*   read    (size_t length)
    {
        if (length >= BLOCK_SIZE)
        {
            length = BLOCK_SIZE - 1;
        }

        if (_index + length >= BLOCK_SIZE)
        {
            size_t i2(0);

            for (size_t i0(_index); i0 < BLOCK_SIZE; ++i0, ++i2)
            {
                _out_buf[i2] = _read_buf[i0];
            }

            int     nread(::read(_file, _read_buf, BLOCK_SIZE));

            if (nread > 0)
            {
                for (size_t i0(0); i2 < length; ++i0, ++i2)
                {
                    _out_buf[i2] = _read_buf[i0];
                }

                _index += length - BLOCK_SIZE;

                _block_start += BLOCK_SIZE;
            }
            else
            {
                _fail = true;
            }
        }
        else
        {
            for (size_t i0(_index), i2(0); i2 < length; ++i0, ++i2)
            {
                _out_buf[i2] = _read_buf[i0];
            }

            _index += length;
        }

        _out_buf[length] = 0;

        return _out_buf;
    }

    inline int   read    ()
    {
        if (_index < BLOCK_SIZE)
        {
            int result(_read_buf[_index]);
            ++_index;
            return result;
        }
        else
        {
            const char* buf(read(1));

            return _fail ? -1  : buf[0];
        }
    }

    inline char*   last ()
    {
        return _out_buf;
    }

protected:
    void set_failed ()
    {
        _fail = true;
    }

private:
    int         _file;
    bool        _fail;
    char        _read_buf[BLOCK_SIZE];  // raw read buffer
    char        _out_buf[BLOCK_SIZE];  // result of read(n)
    size_t      _block_start;
    size_t      _index;
};

#endif // _FILE_READER_H_
